/*	ai.cpp -- Analog Inputs (light sensors, etc)
*
*	Written by Cathy Saxton
*	cathy@idleloop.com
*/

#include <avr\io.h>
#include "abc.h"
#include "debug.h"
#include "ai.h"

/* light sensor readings for black & white */
/* (black values are higher than light values,
   e.g. black = 0xFF, white = 0x7A) */

char AI::m_fInit = fFalse;

/* NOTE: ENUM/LIST #defines are in iop.h */
const IOP AI::m_mpaipiop[aipMax] =
{
	AIP_LIST
};

AI::AI()
{
	if (!m_fInit)
		Init();
}

/* set direction (as input) */
void AI::Init() const
{
	const IOP *piop, *piopMax;

/* set direction for inputs; this uses DDRx sfr's
   and sets bits to 0 to indicate input */
/* REVIEW: optimize would figure out specifics for this prog
   & do them all at once... */
	for (piopMax = (piop = m_mpaipiop) + aipMax; piop < piopMax; ++piop)
		piop->SetAsInput();

	/* enable ADC for reading sensors */
	/* REVIEW: Vref... */
//	outb(ADMUX, 0xC2);	// 1100 0010 (2.56Vref, no L-shift, ADC2)
#ifdef OLD
	outb(ADMUX, 0xE0);	// 1110 0000 (2.56Vref, L-shift, ADC0)
	sbi(ADCSRA, ADEN);	// ADC enable
#endif

	ADMUX = 0xE0;	// 1110 0000 (2.56Vref, L-shift, ADC0)
	ADCSRA = 0x80;	// 1000 0000 (enable, no auto trigger, no interrupt)
	/* use SFIOR to set trigger soure if auto-trigger */

	/* read once since first read takes a while...
	   (which one is irrelevant) */
	(void)Read(aipLsCenter);

	m_fInit = fTrue;
}

/* read sensor attached to analog input aip, returning the result */
uchar AI::Read(uchar aip) const
{
	uchar adc;
	uchar b;

	adc = m_mpaipiop[aip].bit;

	ADMUX &= ~0x1F;		// clear channel selection (low 5 bits)
	ADMUX |= adc;		// select specified channel

#ifdef NO
	/* REVIEW: don't need this?! */
	/* make sure it's not busy... */
	while (bit_is_set(ADCSRA, ADSC))
		;
#endif

	sbi(ADCSRA, ADSC);	// ADC start conversion
		/* (this appears to clear ADIF) */

#ifdef NO
	while (bit_is_clear(ADCSRA, ADSC))	// wait until registering set!
		;
#endif

	/* wait for conversion to complete */
	while (bit_is_clear(ADCSRA, ADIF))
//	while (bit_is_set(ADCSRA, ADSC))
		;

{	// try again...
	sbi(ADCSRA, ADSC);	// ADC start conversion
	while (bit_is_clear(ADCSRA, ADIF))
//	while (bit_is_set(ADCSRA, ADSC))
		;
}

	b = ADCH;

//	sbi(ADCSRA, ADIF);	// clear interrupt flag (yes, by writing 1)

	return b;
}

/* REVIEW: debug only... */
void AI::ShowVals(const COMM &comm) const
{
	int aip;
	uchar b;

	for (aip = 0; aip < aipMax; ++aip)
	{
		b = Read(aip);
		comm.TxByte(b);
		comm.TxSpc();
	}
	comm.TxCRLF();
}

