/*	di.cpp -- Digital Inputs (Buttons, etc)
*
*	Written by Cathy Saxton
*	cathy@idleloop.com
*/

#include <avr\io.h>
#include "abc.h"
#include "debug.h"
#include "di.h"

char DI::m_fInit = fFalse;

/* NOTE: ENUM/LIST #defines are in iop.h */
const IOP DI::m_mpdipiop[dipMax] =
{
	DIP_LIST
};


DI::DI()
{
	if (!m_fInit)
		Init();
}

/* set direction (as input), pull high (open) */
void DI::Init() const
{
	const IOP *piop, *piopMax;

	for (piopMax = (piop = m_mpdipiop) + dipMax; piop < piopMax; ++piop)
		piop->SetAsInputAndPullHigh();

	m_fInit = fTrue;
}

/* returns true if button was pressed; waits for release
   before returning; note that inputs are pulled high
   when open, closing results in pulling signal low */
char DI::FPressed(uchar dip) const
{
	if (!FClosed(dip))		// no press
		return fFalse;

	/* wait for release... */
	while (FClosed(dip))	// pressed
		;
	return fTrue;
}

/* return true if specified input is high (open switch), else false */
uchar DI::FClosed(uchar dip) const
{
	return m_mpdipiop[dip].FLow();
}
