/*	iop.cpp -- shared Input/Output stuff
*
*	Written by Cathy Saxton
*	cathy@idleloop.com
*/

#include <avr\io.h>
#include "abc.h"
#include "debug.h"
#include "iop.h"

void IOP::SetAsInput() const
{
	switch (reg)
	{
	case regA:
		cbi(DDRA, bit);
		break;
	case regB:
		cbi(DDRB, bit);
		break;
	case regC:
		cbi(DDRC, bit);
		break;
	case regD:
		cbi(DDRD, bit);
		break;
	}
}

/* set direction for inputs; this uses DDRx sfr's
   and sets bits to 0 to indicate input */
/* REVIEW: optimize would figure out specifics for this prog
   & do them all at once... */
/* REVIEW: cbi needed for all inputs (i.e. ADC, too...) */
/* pull input high (via internal resistor);
   used for switches; off = high, on = low */
/* REVIEW: automate? do for all? (enum rang efor sw...) */
/* REVIEW: have mask for each register; use one outb instead of
   multiple sbi/cbi calls */
void IOP::SetAsInputAndPullHigh() const
{
	switch (reg)
	{
	case regA:
		cbi(DDRA, bit);
		sbi(PORTA, bit);
		break;
	case regB:
		cbi(DDRB, bit);
		sbi(PORTB, bit);
		break;
	case regC:
		cbi(DDRC, bit);
		sbi(PORTC, bit);
		break;
	case regD:
		cbi(DDRD, bit);
		sbi(PORTD, bit);
		break;
	}
}

char IOP::FLow() const
{
	switch (reg)
	{
	case regA:
		return bit_is_clear(PINA, bit);
	case regB:
		return bit_is_clear(PINB, bit);
	case regC:
		return bit_is_clear(PINC, bit);
	case regD:
		return bit_is_clear(PIND, bit);
	}
	Assert(fFalse);
	return fFalse;	// non-signaled state seems safest...
}


void IOP::SetAsOutput() const
{
	switch (reg)
	{
	case regA:
		sbi(DDRA, bit);
		break;
	case regB:
		sbi(DDRB, bit);
		break;
	case regC:
		sbi(DDRC, bit);
		break;
	case regD:
		sbi(DDRD, bit);
		break;
	}
}

void IOP::SetAsOutputAndSignalHigh() const
{
	switch (reg)
	{
	case regA:
		sbi(DDRA, bit);
		sbi(PORTA, bit);
		break;
	case regB:
		sbi(DDRB, bit);
		sbi(PORTB, bit);
		break;
	case regC:
		sbi(DDRC, bit);
		sbi(PORTC, bit);
		break;
	case regD:
		sbi(DDRD, bit);
		sbi(PORTD, bit);
		break;
	}
}

/* write 1 to this output */
void IOP::SignalHigh() const
{
	switch (reg)
	{
	case regA:
		sbi(PORTA, bit);
		break;
	case regB:
		sbi(PORTB, bit);
		break;
	case regC:
		sbi(PORTC, bit);
		break;
	case regD:
		sbi(PORTD, bit);
		break;
	}
}

/* write 0 to this output */
void IOP::SignalLow() const
{
	switch (reg)
	{
	case regA:
		cbi(PORTA, bit);
		break;
	case regB:
		cbi(PORTB, bit);
		break;
	case regC:
		cbi(PORTC, bit);
		break;
	case regD:
		cbi(PORTD, bit);
		break;
	}
}
