/*	main.cpp
*
*	Written by Cathy Saxton
*	cathy@idleloop.com
*
*	Maze solving program
*
*	Black button will meter for black.
*	White button will meter for white.
*	Reset button will show readings from the 5 light sensors.
*
*	Hit Start to start the line following... The first thing it will
*	do is to help find the line -- when the LED comes on, it's on
*	the line. Hit Start again to follow it!
*
*	For AVR1.1 kit from barello.net
*/

/* REVIEW: figure out how to get make to
   put stuff in sub-dirs... (also #ifdef DEBUG...) */

/* REVIEW
	- make DO, AI, MTR, etc global (saves stack space when need in mult func...)
	- nicer: put them in struct, pass struct...
	- do_ is a hack...
*/

#include <avr\io.h>
#include "abc.h"
#include "debug.h"
#include "di.h"
#include "ai.h"
#include "do.h"
#include "mtr.h"
#include "comm.h"
#include "eeprom.h"
#include "eyes.h"
#include "timer.h"
#include "maze.h"

/* note: ints are 16 bits on Atmel ATMega 16 */
/* int: -32768 to 32767 */
/* ulong: 0 to 4G... (4,294,967,295) */

/*	BSS: currently 0x1aa (was 199 w/out stack comms...)
	stack: 0x419 in find jct
	200 - 400 is 512 bytes!
*/

enum	// ACTions, selected with RESET, run with START
{
	actMin = 1,
	actSolveMaze = actMin,
	actReadings,
	actMax
};

void ResetMaze(const DI &di, const DO &do_, const EEPROM &eeprom);
void SolveMaze(const DI &di, const AI &ai, const DO &do_,
			   const MTR &mtr, const COMM &comm, const EEPROM &eeprom);
void DetectLine(const DI &di, const AI &ai, const DO &do_, EYES *peyes);
uchar GrfPathFindJct(const DI &di, const AI &ai, const MTR &mtr,
					 const TIMER &timer, EYES *peyes, ushort ctickStop);
char FTurn(const DI &di, const AI &ai, const MTR &mtr, const TIMER &timer,
		   MAZE *pmaze, EYES *peyes, uchar grfpath, ushort ctickStop);

#ifdef NO
void Test(const DI &di, const AI &ai, const DO &do_,
		  const MTR &mtr, const COMM &comm, const EEPROM &eeprom)
{
	/* second was good */
	/* write took ~3 seconds */
	ushort ctick;
	TIMER timer;
	ushort e;

	ctick = timer.CtickPlusSec(60);

	while (timer.CtickCur() < ctick)
	{
		ai.ShowVals(comm);
	}
	do_.Blink(dopLedGreen, 3);
	for (e = ecbhMaze; e < 500; ++e)
		eeprom.Write(e, 0x55);
	do_.Blink(dopLedGreen, 3);
}
#endif

int main()
{
	/* REVIEW: temp workaround to disable JTAG
	   (so can use PC3&4 for motors)
	   NOTE: must send command twice! */
//	sbi(MCUCSR, JTD);
//	sbi(MCUCSR, JTD);

	DO do_;		// Digital Outputs

	do_.On(dopLedGreen);

	DI di;		// Digital Inputs
	AI ai;		// Analog Inputs
	MTR mtr;	// MoToRs
	COMM comm;	// COMMunication (serial)
	EEPROM eeprom;
	char act;

	comm.TxSzCRLF("START");
	ai.ShowVals(comm);	// REVIEW: debug only... (all calls to this...)

	do_.Off(dopLedGreen);

	do_.Blink(dopLedGreen, 1);		// signal we're ready

#ifndef INFO
	comm.TxSz("stack: ");
	comm.TxByteHex(SPH);
	comm.TxByteHex(SPL);
	comm.TxCRLF();
	/* BSS... approx 0-0x24b; max stack = 45f */
	/* BSS... 0x199 */
#endif

	act = actMin;

	while (fTrue)	// react to switch presses
	{
		if (di.FPressed(dipBtnStart))
		{
			comm.TxSzCRLF("run");
			switch (act)
			{
			case actSolveMaze:
				SolveMaze(di, ai, do_, mtr, comm, eeprom);
				break;
			case actReadings:
				ai.ShowVals(comm);
				break;
			}
		}
		if (di.FPressed(dipBtnBlack))
		{
			EYES eyes;

			comm.TxSzCRLF("meter black");
			ai.ShowVals(comm);
			eyes.MeterBlack(ai);
			do_.Blink(dopLedGreen, 2);
		}
		if (di.FPressed(dipBtnWhite))
		{
			EYES eyes;

			comm.TxSzCRLF("meter white");
			ai.ShowVals(comm);
			eyes.MeterWhite(ai);
			do_.Blink(dopLedGreen, 2);
		}
		if (di.FPressed(dipBtnReset))
		{
			ResetMaze(di, do_, eeprom);
#ifdef OLD
			if (++act >= actMax)
				act = actMin;
			comm.TxSz("act = ");
			comm.TxNum(act);
			comm.TxCRLF();
			do_.Blink(dopLedGreen, act);
//			ai.ShowVals(comm);
#endif
		}
	}

	return 0;
}

#define csecReset	5
#define cpressReset	3
void ResetMaze(const DI &di, const DO &do_, const EEPROM &eeprom)
{
	TIMER timer;
	MAZE maze;
	char c;
	ushort ctick;

	c = 0;
	do_.On(dopLedGreen);
	/* have to press start cpressReset times in csecReset seconds */
	ctick = timer.CtickPlusSec(csecReset);
	while (timer.CtickCur() < ctick && c < cpressReset)
//	while (c < cpressReset)
	{
		if (di.FPressed(dipBtnStart))
			++c;
		if (di.FPressed(dipBtnReset))
			break;
	}
	do_.Off(dopLedGreen);
	if (c >= cpressReset)
	{
		maze.Clear(eeprom);
		do_.Blink(dopLedGreen, 4);
	}
}


#define secSolve (3*60-5)	// how many seconds to spend solving the maze (3 minutes)

#define dlsvTolerance 5	// tolerance acceptable between line sensor readings
						//     when following line

//#define FAST
#ifdef FAST
#define spdBase 0xF0	// general run speed
#define spdTurn 0xC0	// speed when turning at jct
#else // FAST
#define spdBase 0xC0	// general run speed
#define spdTurn 0x80	// speed when turning at jct
#endif // FAST

#define factStayOnLine	0x01
#define factCheckL		0x02
#define factCheckR		0x04

/* we move wheels forward so they're even with side junctions
   when examining the junction (for paths); that makes calculating
   distance between junctions easier */
#ifdef OPTICAL_DIST
#define distCheckJct	192
//#define distCheckJct	150
//#define distCheckJct	120
#else // OPTICAL_DIST
/* (using time/ticks to measure distance) */
#define distCheckJct	52	// works well with spdBase = 0xF0, 0xC0
#endif // OPTICAL_DIST
/* this is the distance between where the sensors stop and where the wheel
   axis is; it's used for calculating distance traveled; we start that
   far off the line AFTER A TURN, and want to move that far
   before executing a turn */

void SolveMaze(const DI &di, const AI &ai, const DO &do_,
			   const MTR &mtr, const COMM &comm, const EEPROM &eeprom)
{
	TIMER timer;
	MAZE maze;
	EYES eyes;
	uchar grfpath;
	ushort ctickStop;
	char fStop;

#ifdef INFO
	comm.TxSz("stack: ");
	comm.TxByteHex(SPH);
	comm.TxByteHex(SPL);
	comm.TxCRLF();
#endif

	maze.Load(eeprom);

	DetectLine(di, ai, do_, &eyes);
	ctickStop = timer.CtickPlusSec(secSolve);

	fStop = fFalse;
	do
	{
		grfpath = GrfPathFindJct(di, ai, mtr, timer, &eyes, ctickStop);
		if (grfpath & fpathStop)
		{
			fStop = fTrue;
		}
		else
		{
//			dist = mtr.Distance();
//			do_.Blink(dopLedGreen, 1);
			if (!FTurn(di, ai, mtr, timer, &maze, &eyes, grfpath, ctickStop))
				fStop = fTrue;
		}
	}
	while (!fStop);
	do_.Blink(dopLedGreen, 1);
	maze.CleanupAndSave(eeprom);		// clean up any loose ends, store to EEPROM
	do_.Blink(dopLedGreen, (grfpath & fpathEnd) ? 3 : 1);
//	while (!di.FPressed(dipBtnStart))	// REVIEW
		;
//	maze.Show();
}

/* check for line under middle sensor; turn on LED when line found,
   off when lost; exit when Start button pressed */
void DetectLine(const DI &di, const AI &ai, const DO &do_, EYES *peyes)
{
	int fOnLine;

	fOnLine = fFalse;
	while (!di.FPressed(dipBtnStart))
	{
//		peyes->Read(lsCenter, ai);
		if (fOnLine)
		{
			if (peyes->FLostLine(lsCenter, ai))
			{
				fOnLine = fFalse;
				do_.Off(dopLedGreen);
			}
		}
		else if (peyes->FFoundLine(lsCenter, ai))
		{
			fOnLine = fTrue;
			do_.On(dopLedGreen);
		}
	}
	do_.Off(dopLedGreen);	// turn off LED since we're done using it
}


/* drive forward until find junction (dead end, left/right turn,
   or intersection (plus or T)); stops motors */
uchar GrfPathFindJct(
	const DI &di,		// digital inputs (buttons)
	const AI &ai,		// analog inputs (light sensors)
	const MTR &mtr,		// motors (including distance measurement)
	const TIMER &timer,
	EYES *peyes,		// group of light sensors; used to detect lines
	ushort ctickStop)	// when we have to stop
{
	uchar grfact;
	uchar spdL, spdR;
	short dlsv;
//	TIMER timer;
	Debug(COMM comm;)
	short distMax;
	uchar grfpath;

#ifdef INFO
	comm.TxSz("stack [find jct]: ");
	comm.TxByteHex(SPH);
	comm.TxByteHex(SPL);
	comm.TxCRLF();
#endif

	grfact = factStayOnLine | factCheckL | factCheckR;
	grfpath = 0;
	distMax = 0;
	spdL = spdR = spdBase;
	mtr.ClearDistance();
	mtr.Fwd(spdBase);
	do
	{
		if (timer.CtickCur() >= ctickStop)
		{
			grfpath = fpathStop | fpathTime;
			goto LDone;
		}
		if (di.FPressed(dipBtnStart))
		{
			grfpath = fpathStop | fpathUser;
			goto LDone;
		}

		if (grfact & factStayOnLine)
		{
			peyes->Read(lsLeft, ai);
			peyes->Read(lsCenter, ai);
			peyes->Read(lsRight, ai);
			if (peyes->FLostLine(lsCenter) && peyes->FLostLine(lsLeft)
				&& peyes->FLostLine(lsRight))
			{
				/* end of line (forward) */
				grfact &= ~factStayOnLine;
				if (distMax == 0)
				{
					/* haven't found any paths to side; this is a dead end
					   (so, we don't need to move forward to check jct
					   and set up for distance measurement) */
					comm.TxCh('X');
					comm.TxCRLF();
					Assert(grfpath == 0);
					goto LDone;
				}
				else
				{
					comm.TxCh('-');
					comm.TxCRLF();
					/* still checking sides; go roughly straight */
					mtr.Fwd(spdBase);
				}
			}
			else
			{
				dlsv = peyes->Dlsv(lsRight, lsLeft);
				/* cast to ushort so doesn't sign extend!!! */
//				dlsr = (short)((ushort)mpposlslsr[posCur][lsRight])
//					- (short)((ushort)mpposlslsr[posCur][lsLeft]);
				if (dlsv > dlsvTolerance)
				{
					/* too far L (need to turn R) */
					/* R needs to go slower... */
					spdL = spdBase;
					/* REVIEW: make sure not neg!!! */
					spdR = (uchar)(spdBase - (dlsv * 2));
					/* the farther off, the slower it goes (meaning
					   a bigger difference in speed & faster turning) */
					mtr.Fwd(spdL, spdR);
				}
				else if (dlsv < -dlsvTolerance)
				{
					/* too far R (need to turn L) */
					/* L needs to go slower... */
					spdR = spdBase;
					/* REVIEW: make sure not neg!!! */
					spdL = (uchar)(spdBase + (dlsv * 2)); // dlsr neg, so this slows down motor...
					/* the farther off, the slower it goes (meaning
					   a bigger difference in speed & faster turning) */
					mtr.Fwd(spdL, spdR);
				}
				else if (spdL != spdR)
				{
					spdL = spdR = spdBase;
					mtr.Fwd(spdL, spdR);
				}
			}
		}

		if (grfact & factCheckL)
		{
			if (peyes->FFoundLine(lsFarLeft, ai))
			{
				/* check for false positive... */
				mtr.Off();
				comm.TxCh('?');
				comm.TxCh('L');
				comm.TxNum(peyes->m_dnls[lsFarLeft].alsv[ilsvCur]);
				if (peyes->FFoundLine(lsFarLeft, ai))
				{
					/* OK, looks legit */
					grfpath |= fpathL;
					grfact &= ~factCheckL;
					goto LCheckJct;
				}
				comm.TxCRLF();
				mtr.On();	// resume...
			}
		}

		if (grfact & factCheckR)
		{
			if (peyes->FFoundLine(lsFarRight, ai))
			{
				/* check for false positive... */
				mtr.Off();
				comm.TxCh('?');
				comm.TxCh('R');
				comm.TxNum(peyes->m_dnls[lsFarRight].alsv[ilsvCur]);
				if (peyes->FFoundLine(lsFarRight, ai))
				{
					/* OK, looks legit */
					grfpath |= fpathR;
					grfact &= ~factCheckR;
LCheckJct:
					comm.TxCh('Y');
					/* move forward to continue checking jct (for other
					   side, straight); this also moves fwd so wheels
					   are even with side path(s) for easy turning
					   and distance measurement */
					if (distMax == 0)
						distMax = mtr.Distance() + distCheckJct;
				}
				comm.TxCRLF();
				mtr.On();	// resume...
			}
		}

		if (distMax != 0 && mtr.Distance() >= distMax)
		{
			/* done checking junction */
			mtr.Off();
			comm.TxCh('*');
			comm.TxCRLF();
			if (grfact & factStayOnLine)	// still on path
			{
				grfpath |= fpathC;
				if ((grfact & (factCheckL | factCheckR)) == 0)
				{
					/* found L & R; see if still have 'em... */
					if (peyes->FFoundLine(lsFarLeft, ai)
						&& peyes->FFoundLine(lsFarRight, ai))
					{
						/* double-check! */
						if (peyes->FFoundLine(lsFarLeft, ai)
							&& peyes->FFoundLine(lsFarRight, ai))
						{
							grfpath = fpathStop | fpathEnd;
						}
					}
				}
			}
			goto LDone;
		}

	}
	while (true);	// REVIEW: fix this!

LDone:
	mtr.Off();
	return grfpath;
}

/* return false if time or user halt */
char FTurn(
	const DI &di,
	const AI &ai,
	const MTR &mtr,
	const TIMER &timer,
	MAZE *pmaze,
	EYES *peyes,
	uchar grfpath,
	ushort ctickStop)	// when we have to stop
{
	COMM comm;
	char iMax;
	char f;

#ifdef INFO
	comm.TxSz("stack [turn]: ");
	comm.TxByteHex(SPH);
	comm.TxByteHex(SPL);
	comm.TxCRLF();
	/* BSS... approx 0-0x24b; max stack = 45f */
	/* BSS... 0x199 */
#endif

	iMax = 1;
	comm.TxSz("turn: ");
	switch (pmaze->TurnAtJct(mtr.Distance(), grfpath))
	{
	case turn180Special:
		iMax = 2;
		comm.TxCh('.');
		/* fall through */
	case turn180:
		comm.TxCh('U');
		mtr.Right(spdTurn);
		break;
	case turnR:
		comm.TxCh('R');
		mtr.Right(spdTurn);
		break;
	case turnL:
		comm.TxCh('L');
		mtr.Left(spdTurn);
		break;
	case turnNone:
	default:
		comm.TxCh('S');
		/* yippee; we're done! */
		iMax = 0;
		break;
	}
	comm.TxCRLF();
//	pmaze->Show();

	while (--iMax >= 0)
	{
		/* move off line */
		mtr.On();	// this is OK even if we're already on, and needed for iMax>1
		do
		{
			f = peyes->FLostLine(lsCenter, ai);
			if (f)
			{
				/* verify! */
				mtr.Off();
				f = peyes->FLostLine(lsCenter, ai);
				if (!f)
					mtr.On();
			}
		}
		while (!f);
		/* turn until find line */
		mtr.On();
		do
		{
			if (timer.CtickCur() >= ctickStop)
			{
				grfpath = fpathStop | fpathTime;
				goto LDone;
			}
			if (di.FPressed(dipBtnStart))
			{
				grfpath = fpathStop | fpathUser;
LDone:
				mtr.Off();
				return fFalse;
			}
			f = peyes->FFoundLine(lsCenter, ai);
			if (f)
			{
				/* verify! */
				mtr.Off();
				f = peyes->FFoundLine(lsCenter, ai);
				if (!f)
					mtr.On();
			}
		}
		while (!f);
	}

//	while (!di.FPressed(dipBtnStart))	// REVIEW: temp...
		;
	return fTrue;
}

#ifdef READ_SENSORS
			uchar alsv[20][aipMax];
			int i, aip;

			for (i = 0; i < 20; ++i)
			{
				for (aip = 0; aip < aipMax; ++aip)
					alsv[i][aip] = ai.Read(aip);
			}
			for (i = 0; i < 20; ++i)
			{
				for (aip = 0; aip < aipMax; ++aip)
				{
					comm.TxByte(alsv[i][aip]);
					comm.TxCh(' ');
				}
				comm.TxCRLF();
			}
#endif
#ifdef OPTICAL_TEST
			TIMER timer;

			mtr.ClearDistance();
			mtr.Fwd(0xC0);
			timer.Wait(2);
			mtr.Off();
			comm.TxSz("fwd: ");
			comm.TxNumSigned(mtr.Distance());

			mtr.ClearDistance();
			mtr.Rev(0xC0);
			timer.Wait(2);
			mtr.Off();
			comm.TxSz("   rev: ");
			comm.TxNumSigned(mtr.Distance());
			comm.TxCRLF();
#endif

#ifdef INFO
	comm.TxSz("stack: ");
	comm.TxByteHex(SPH);
	comm.TxByteHex(SPL);
	comm.TxCRLF();
#endif

#ifdef NO
	while (!di.FPressed(dipBtnStart))
	{
		lsrL = ai.Read(aipLsLeft);
		lsrC = ai.Read(aipLsCenter);
		lsrR = ai.Read(aipLsRight);
		comm.TxSz("L: ");
		comm.TxByte(lsrL);
		comm.TxSz("  C: ");
		comm.TxByte(lsrC);
		comm.TxSz("  R: ");
		comm.TxByte(lsrR);
		comm.TxCRLF();
		timer.WaitTicks(16);
	}
#endif

